"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deserializeLayout = void 0;
const layout_1 = require("./layout");
const utils_1 = require("./utils");
function deserializeLayout(layout, encoded, offset = 0, consumeAll = true) {
    const [decoded, finalOffset] = internalDeserializeLayout(layout, encoded, offset);
    if (consumeAll && finalOffset !== encoded.length)
        throw new Error(`encoded data is longer than expected: ${encoded.length} > ${finalOffset}`);
    return consumeAll ? decoded : [decoded, finalOffset];
}
exports.deserializeLayout = deserializeLayout;
function internalDeserializeLayout(layout, encoded, offset) {
    let decoded = {};
    for (const item of layout)
        [(item.omit ? {} : decoded)[item.name], offset] =
            deserializeLayoutItem(item, encoded, offset);
    return [decoded, offset];
}
function updateOffset(encoded, offset, size) {
    const newOffset = offset + size;
    if (newOffset > encoded.length)
        throw new Error(`encoded data is shorter than expected: ${encoded.length} < ${newOffset}`);
    return newOffset;
}
function deserializeUint(encoded, offset, size) {
    let value = 0n;
    for (let i = 0; i < size; ++i)
        value += BigInt(encoded[offset + i]) << BigInt((size - 1 - i) * 8);
    return [
        ((size > layout_1.numberMaxSize) ? value : Number(value)),
        updateOffset(encoded, offset, size)
    ];
}
function deserializeLayoutItem(item, encoded, offset) {
    try {
        switch (item.binary) {
            case "uint": {
                const [value, newOffset] = deserializeUint(encoded, offset, item.size);
                if ((0, layout_1.isUintType)(item.custom)) {
                    (0, utils_1.checkUintEquals)(item.custom, value);
                    return [item.custom, newOffset];
                }
                if ((0, layout_1.isUintType)(item?.custom?.from)) {
                    (0, utils_1.checkUintEquals)(item.custom.from, value);
                    return [item.custom.to, newOffset];
                }
                return [
                    item.custom !== undefined ? item.custom.to(value) : value,
                    newOffset
                ];
            }
            case "bytes": {
                let newOffset;
                let fixedFrom;
                let fixedTo;
                if (item.custom !== undefined) {
                    if ((0, layout_1.isBytesType)(item.custom))
                        fixedFrom = item.custom;
                    else if ((0, layout_1.isBytesType)(item.custom.from)) {
                        fixedFrom = item.custom.from;
                        fixedTo = item.custom.to;
                    }
                }
                if (fixedFrom !== undefined)
                    newOffset = updateOffset(encoded, offset, fixedFrom.length);
                else {
                    item = item;
                    if ("size" in item && item.size !== undefined)
                        newOffset = updateOffset(encoded, offset, item.size);
                    else if (item.lengthSize !== undefined) {
                        let length;
                        [length, offset] = deserializeUint(encoded, offset, item.lengthSize);
                        newOffset = updateOffset(encoded, offset, length);
                    }
                    else
                        newOffset = encoded.length;
                }
                const value = encoded.slice(offset, newOffset);
                if (fixedFrom !== undefined) {
                    (0, utils_1.checkUint8ArrayDeeplyEqual)(fixedFrom, value);
                    return [fixedTo ?? fixedFrom, newOffset];
                }
                return [
                    item.custom !== undefined ? item.custom.to(value) : value,
                    newOffset
                ];
            }
            case "array": {
                let ret = [];
                const { arrayItem } = item;
                const deserializeArrayItem = (index) => {
                    const [deserializedItem, newOffset] = deserializeLayoutItem({ name: `${item.name}[${index}]`, ...arrayItem }, encoded, offset);
                    ret.push(deserializedItem);
                    offset = newOffset;
                };
                if (item.lengthSize !== undefined) {
                    const [length, newOffset] = deserializeUint(encoded, offset, item.lengthSize);
                    offset = newOffset;
                    for (let i = 0; i < length; ++i)
                        deserializeArrayItem(i);
                }
                else
                    while (offset < encoded.length)
                        deserializeArrayItem(ret.length);
                return [ret, offset];
            }
            case "object": {
                return internalDeserializeLayout(item.layout, encoded, offset);
            }
            case "switch": {
                const [id, newOffset] = deserializeUint(encoded, offset, item.idSize);
                const { idLayoutPairs } = item;
                if (idLayoutPairs.length === 0)
                    throw new Error(`switch item '${item.name}' has no idLayoutPairs`);
                const hasPlainIds = typeof idLayoutPairs[0][0] === "number";
                const pair = idLayoutPairs.find(([idOrConversionId]) => hasPlainIds ? idOrConversionId === id : (idOrConversionId)[0] === id);
                if (pair === undefined)
                    throw new Error(`unknown id value: ${id}`);
                const [idOrConversionId, idLayout] = pair;
                const [decoded, nextOffset] = internalDeserializeLayout(idLayout, encoded, newOffset);
                return [
                    { [item.idTag ?? "id"]: hasPlainIds ? id : idOrConversionId[1],
                        ...decoded },
                    nextOffset
                ];
            }
        }
    }
    catch (e) {
        e.message = `when deserializing item '${item.name}': ${e.message}`;
        throw e;
    }
}
//# sourceMappingURL=deserialize.js.map